#version 330

// Load vertex, normals, texcoords and tangets:
layout (location = 0) in vec3 vPosition;
layout (location = 1) in vec3 vNormal;
layout (location = 2) in vec3 vTangent;
layout (location = 3) in vec2 vUV;

// Output UV-Coordinates to render the texture: 
out vec3 position;
out vec2 uv;
out vec3 normal;
out vec3 tangent;

out mat3 TBN;

/**
 * Projection matrix which performs a perspective transformation.
 */
uniform mat4 projection;

/**
 * View matrix which transforms the vertices from world space
 * into camera space.
 */
uniform mat4 view;

/**
 * Model matrix which transforms the vertices from model space
 * into world space.
 */
uniform mat4 model;

/**
 * Since this is the vertex shader, the following program is
 * executed for each vertex (i.e. triangle point) individually.
 */
void main()
{
    mat4 modelView = view * model;

    // Calculate position in camera space:
    vec4 position4f = modelView * vec4(vPosition, 1.0);
    position = position4f.xyz;

    // Set uv coordinate:
    uv = vUV;

    // Calculate Normal matrix:
    mat3 normalMatrix = transpose(inverse(mat3(modelView)));

    vec3 N = normalMatrix * vNormal;
    vec3 T = normalMatrix * vTangent;
    vec3 B = cross(N, T);
    TBN = mat3(T, B, N);

    // Calculate normal in camera space:
    normal = normalize(N);
    tangent = normalize(T);

    // Project position:
    gl_Position = projection * position4f;
}
